/*
Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* System includes */
#include <time.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/timex.h>
#include <sys/timeb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

/* Our includes */
#include "flow_stats.h"
#include "conn_stats.h"
#include "hash.h"
#include "dward.config"
#include "classifier.h"
#include "rate_limit.h"
#include "process.h"

/* Semaphores used for exclusive access */
extern pthread_mutex_t flow_mutex;
extern pthread_mutex_t conn_mutex;

/* Flow hash */
extern struct hash_table* FlowHash; 

/* Connection hash */
extern struct hash_table* ConnHash; 

/* Debug flag */
extern int DEBUG;

/* INC flag */
extern int INC;

/* Debug files */
extern FILE* flowdebug;
extern FILE* conndebug;

/* Function that prints packet statistics to a file 
   for charting */
extern void print_stats(void);

/* We classify connections very frequently so we can 
   quickly differentiate good ones from bad ones. 
   With lower frequency we classify flows and adjust 
   rate limit */
void process()
{
  int count = 0;
  while(TRUE)
    {
      /* Every so often */
      sleep(CONN_OBS_INTERVAL);
      process_conns();
      count += CONN_OBS_INTERVAL;
      if (count >= FLOW_OBS_INTERVAL)
	{
	  process_flows();
	  count = 0;
	}
      print_stats();
    }
}


/* Classify and rate-limit flows */
void process_flows()
{
  long i, curtime;

  /* Obtain lock */
  curtime = time(0);
  for(i=0;i<FlowHash->size;i++)
    {
      /* For each valid flow */
      if(FlowHash->entries[i].state == VALID)
	{
	  /* Classify flow based on model */
	  switch(FLOW_MODEL)
	    {
	    case SIMPLE_TH: 
	      threshold_classify(&FlowHash->entries[i]);
	      break;
	    case MULTI_PARAM: 
	      multiparam_classify(&FlowHash->entries[i]);
	      break;
	    case HMM: 
	      hmm_classify(&FlowHash->entries[i]);
	      break;
	    default: break;
	    }
	  /* Define rate limit */
	  switch(RATE_LIMIT)
	    {
	    case FIX_RATE_LIMIT: 
	      break;
	    case VAR_RATE_LIMIT:
	      variable_rate_limit(&FlowHash->entries[i]);
	      break; 
	    case PERCENT: 
		  break;
	    default: break;
	    }
	  /* Now we can reset statistics and erase drops for new round;
	     drop erase will also propagate to kernel */
	  reset_flow(i);
	}
    }
}

/* Classify connections 
   and adjust good-connection list  */
void process_conns()
{
  long i, curtime;

  curtime = time(0);
  for(i=0;i<ConnHash->size;i++)
    {
      /* For each valid connection */
      if(ConnHash->entries[i].state == VALID)
	{
	  /* Classify connection based on model */
	  switch(CONN_MODEL)
	    {
	    case PROTOCOL: 
	      protocol_classify(&ConnHash->entries[i]);
	      break;
	    default: 
	      break; 
	   }
	  /* Now we can reset statistics and erase drops for new round;
	     drop erase will also propagate to kernel */
	  reset_connection(i);
	}
    }
}
















