/* Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* System includes */
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

/* Our includes */
#include "hash.h"
#include "rl_stats.h"
#include "dward.config"
#include "dward.h"

/* Hash for rate-limited flows 
   This variable could have been joined to 
   flow stats but we assume that number of 
   rate-limited flows will be much smaller than
   number of all flows so separate hash is 
   better choice */
struct hash_table* RateLimitHash; 

/* Some thread related variables */
pthread_mutex_t ratelimit_mutex = PTHREAD_MUTEX_INITIALIZER; 

/* Debug file for rate limit*/
FILE* rldebug;

/* Debug flag */
extern int DEBUG;

/* Hash related functions we need to define */
int rl_clean(struct rl_stats* data)
{
  /* Rate limit entry is only deleted
     explicitly */
    return 0;
}

int rl_compare(unsigned int* k1, unsigned int* k2)
{
  if (k1 == NULL || k2 == NULL)
    return FALSE;

  if (*k1==*k2)
    return 1;
  else
    return 0;
}

unsigned long rl_hash1(unsigned int* k, unsigned long size)
{
  return ((*k) % size);
}

unsigned long rl_hash2(unsigned int* k, unsigned long size)
{
  return (1+(*k)%(size-1));
}

/* End of required hash related functions */

/* Finds entry for given IP address in hash, 
   really just masks the fact that 
   we use hash for statistics bookkeeping 
   input: IP address in char array (dotted notation)			       
   output: rate limit (-1 not found)	*/
int find_rate_limit(char *address)
{
  struct in_addr in;  
  unsigned long num;
  inet_aton(address,&in);
  num = in.s_addr;
  return find_rate_limit_num(num); 
}

/* Finds entry for given IP address in numeric form, 
   really just masks the fact that 
   we use hash for statistics bookkeeping 
   input: unsigned int IP address 			       
   output: index (-1 not found)	*/
int find_rate_limit_num(unsigned int address)
{
  long index;
  struct rl_stats* rs;
  int rate;

  index = find(&address,RateLimitHash);

  if (index == NOT_FOUND)
    rate = -2;
  else
    {
      rs = ((struct rl_stats*)RateLimitHash->entries[index].data);
      rate = rs->rate_limit;
    }
  return rate;
}

/* Add IP address to hash, return index 
   input: IP address in char array (dotted notation)			       
   output: index */ 
long add_rate_limit(char* address)
{
  struct in_addr in;
  unsigned long num;

  inet_aton(address,&in);
  num = in.s_addr;
  return add_rate_limit_num(num);
}

/* Add IP address to hash, return index 
   input: unsigned int IP address			       
   output: index */ 
long add_rate_limit_num(unsigned int address)
{
  long index;
  int outcome  = 0;
  struct rl_stats* data;
  index = insert(&address,RateLimitHash,&outcome);

  if (index == NOT_FOUND)
    return NOT_FOUND;
  if(outcome == EXISTING)
    return index;

  data = (struct rl_stats*)RateLimitHash->entries[index].data;

  /* create and initialize flow structure */
  data->rate_limit = -1;


  if (DEBUG > 1)
    {
      unsigned long curtime = time(0);
      struct in_addr in;
      in.s_addr = address;

      if (index != NOT_FOUND)
	  fprintf(rldebug,
		  "%ld: Inserted address %s at index %ld\n",
		  curtime,inet_ntoa(in),index);
      else
	  fprintf(rldebug,
		  "%ld: Could not insert address %s\n",
		  curtime,inet_ntoa(in));
      fflush(rldebug);
    }
  return index;
}

/* When rate limit component decides to modify rate limit
   it invokes update_rate_limit.
   input: IP address in char array (dotted notation),
          int desired rate limit */
void update_rate_limit(char* address, int rate_limit)
{
  struct in_addr in;  
  unsigned long num;
  inet_aton(address,&in);
  num = in.s_addr;
  update_rate_limit_num(num, rate_limit);
}

/* Updates rate limit for given address
   input: unsigned int IP address
          int desired rate limit*/
void update_rate_limit_num(unsigned int address, int rate_limit)
{
  long index;
  struct rl_stats* rl;
  int outcome = 0;
  index = insert(&address,RateLimitHash, &outcome);

  if (index == NOT_FOUND)
    return NOT_FOUND;

  /* If entry does not exist, add it */
  if (outcome == NEW)
    {
      struct rl_stats* data;

      data = (struct rl_stats*)RateLimitHash->entries[index].data;
      
      /* create and initialize flow structure */
      data->rate_limit = rate_limit;
	    
      if (DEBUG > 1)
	{
	  unsigned long curtime = time(0);
	  struct in_addr in;
	  in.s_addr = address;

	  if (index != NOT_FOUND)
	    fprintf(rldebug,
		    "%ld: Inserted address %s at index %ld\n",
		    curtime,inet_ntoa(in),index);
	  else
	    fprintf(rldebug,
		    "%ld: Could not insert address %s\n",
		    curtime,inet_ntoa(in));
	  fflush(rldebug);
	}
    }
  else
    {
      /* Flow already exists in rate-limit hash, just update the rate */
      rl = (struct rl_stats*)(RateLimitHash->entries[index].data);
      rl->rate_limit = rate_limit;
    }
  
  /* System call to update rate limit in kernel, too */ 
  addflow(address,rate_limit);
  fflush(stdout);
}

/* Delete existing rate limit 
   input: IP address in char array (dotted notation) */
void delete_rate_limit(char* address)
{
  struct in_addr in;  
  unsigned long num;
  inet_aton(address,&in);
  num = in.s_addr;
  delete_rate_limit_num(num);
}

/* Delete existing rate limit 
   input: IP address in numeric notation */
void delete_rate_limit_num(unsigned int address)
{
  long index;

  index = find(&address,RateLimitHash);
  /* If entry does not exist, add it */
  if (index == NOT_FOUND)
    return;

  if (DEBUG)
    {
      unsigned long curtime = time(0);
      struct in_addr in;
      in.s_addr = address;

      fprintf(rldebug,
	      "%ld: Deleted rate limit for %s\n",
		  curtime,inet_ntoa(in));
      fflush(rldebug);
    }

  delete(&address,RateLimitHash);
  
  /* Delete rate limit from kernel, too */
  delflow(address);
}

/* Function that deletes hash related variables in case of
   interrupted operaton */
void rlhash_cleanup(int signo)
{
  free(RateLimitHash->entries);
  free(RateLimitHash);
}

/* Initializing flow statistics hash */
void initialize_rl()
{
  /* Create hash with same min and max size to avoid rehashing
     since we did not use locks for any operation */
  create_hash(&RateLimitHash, RL_HASH_SIZE, RL_HASH_SIZE,
	      RL_HASH_SIZE, RL_REHASH_LOAD_FACTOR_HIGH, 
	      RL_REHASH_LOAD_FACTOR_LOW,  
	      sizeof(unsigned int),
	      sizeof(struct rl_stats),
	      (int (*)(void*))rl_clean, 
	      (int (*)(void*,void*))rl_compare, 
	      (unsigned long (*)(void*,unsigned long))rl_hash1,
	      (unsigned long (*)(void*,unsigned long))rl_hash2);

}










